'use strict';

const path = require('path');
const fse = require('fs-extra');
const pkgDir = require('pkg-dir');
const npmInstall = require('npminstall');
const pathExists = require('path-exists');
const utils = require('@foundation-cli/utils');
const {
  getDefaultRegistry,
  getNpmLatestVersion
} = require('@foundation-cli/get-npm-info');
const formatPath = require('@foundation-cli/format-path');

class Package {
  constructor(options) {
    if (!options) {
      throw new Error('Package类的options参数不能为空')
    }
    if (!utils.isObject(options)) {
      throw new Error('Package类的options参数必须是对象类型')
    }
    // package的路径
    this.targetPath = options.targetPath;
    // package的缓存路径, 我们把npm包从远程下载到本地，需要定义一个存储的路径
    this.storePath = options.storePath;
    // package的name
    this.packageName = options.packageName;
    // package的version
    this.packageVersion = options.packageVersion;
    // package缓存目录前缀
    this.cacheFilePathPrefix = this.packageName.replace('/', '_');
  }
  // 本地缓存中的包名并不是我们在镜像中的包名，而是有一定的转化规则的：@foundation-cli/init ${版本号} -> _@foundation-cli_init@${版本号}@@foundation-cli
  get cacheFilePath() {
    return this.getSpecificCacheFilePath(this.packageVersion);
  }

  getSpecificCacheFilePath(version) {
    return path.resolve(this.storePath, `_${this.cacheFilePathPrefix}@${version}@${this.packageName}`)
  }

  async prepare() {
    if (this.storePath && !pathExists.sync(this.storePath)) { // 当this.storePath目录不存在的时候，提前进行创建
      fse.mkdirpSync(this.storePath); // mkdirpSync：将this.storePath整个路径上没有创建好的部分，全部创建好
    }
    // 当用户传入的版本号是latest时，这在镜像中拉代码是可行的，但是如果你要从本地缓存中去寻找包是不行的，还是得用具体的版本号
    if (this.packageVersion === 'latest') {
      this.packageVersion = await getNpmLatestVersion(this.packageName);
    }
  }

  // 判断package是否存在
  async exists() {
    if (this.storePath) {
      await this.prepare();
      return pathExists.sync(this.cacheFilePath);
    } else { // 当用户指定了targetPath时，storePath会为空的（也就是不使用缓存模式），这时候就判断用户指定的targetPath是否存在
      return pathExists.sync(this.targetPath);
    }
  }

  // 安装package
  async install() {
    await this.prepare();
    // 使用 npminstall 模块下载包
    await npmInstall({
      root: this.targetPath,
      storeDir: this.storePath,
      registry: getDefaultRegistry(true),
      pkgs: [
        {
          name: this.packageName,
          version: this.packageVersion
        }
      ]
    });
  }

  // 更新package
  async update() {
    await this.prepare();

    // 第一步：获取最新版本号
    const latestPackageVersion = await getNpmLatestVersion(this.packageName);

    // 第二步：查询最新版本号对应的路径是否存在
    const latestPackagePath = this.getSpecificCacheFilePath(latestPackageVersion);

    // 第三步：如果第二步的路径不存在，则直接安装最新版本
    if (!pathExists.sync(latestPackagePath)) {
      await npmInstall({
        root: this.targetPath,
        storeDir: this.storePath,
        registry: getDefaultRegistry(true),
        pkgs: [
          {
            name: this.packageName,
            version: latestPackageVersion
          }
        ]
      });

      this.packageVersion = latestPackageVersion;
    }
  }

  // 获取入口文件路径
  getRootFilePath() {

    /**
     * 1. 获取package.json的所在路径 - pkg-dir
     * 2. 读取package.json - require()
     * 3. 寻找main/lib - 获取path
     * 4. 路径的兼容（macOS/windows）
     */
    function _getRootPath(targetPath) {
      const dir = pkgDir.sync(targetPath);

      if (dir) {
        const pkgFile = require(path.resolve(dir, 'package.json'));

        if(pkgFile && pkgFile.main) {
          return formatPath(path.resolve(dir, pkgFile.main)); // path.resolve可以进行路径兼容
        }
      }

      return null;
    }

    if (this.storePath) {
      return _getRootPath(this.cacheFilePath);
    } else {
      return _getRootPath(this.targetPath);
    }
  }
}

module.exports = Package;
